import statistics
from typing import Union, List, Optional

import torch.linalg
from torch import Tensor
from torch_geometric.typing import Adj, OptTensor
from torch_geometric.utils import to_dense_adj
from torch_sparse import SparseTensor

from graphesn import DynamicData

__all__ = ['graph_spectral_norm', 'approximate_graph_spectral_radius',
           'compute_dynamic_graph_alpha', 'compute_dynamic_weighted_graph_alpha',
           'distance_to_proximity', 'to_sparse_adjacency']


def graph_spectral_norm(edge_index: Adj, edge_weight: OptTensor = None):
    """
    Spectral norm of a graph

    :param edge_index: Graph adjacency matrix
    :param edge_weight: Edges weight (optional)
    :return: Spectral norm
    """
    return float(torch.linalg.matrix_norm(to_dense_adj(edge_index, edge_attr=edge_weight), ord=2))


def approximate_graph_spectral_radius(adj: SparseTensor, max_iterations: int = 1000, threshold: float = 1e-6):
    """
    Spectral radius of a graph via power method iteration

    :param adj: Sparse adjacency matrix
    :param max_iterations: Maximum number of power iterations
    :param threshold: Convergence threshold for dominant eigenvector
    :return: Spectral radius
    """
    u = torch.rand(adj.size(0), 1).to(adj.device())
    u /= torch.linalg.vector_norm(u)
    for _ in range(max_iterations):
        v = adj.matmul(u)
        v /= torch.linalg.vector_norm(v)
        if torch.linalg.vector_norm(v - u) < threshold:
            break
        else:
            u = v
    alpha = (adj.matmul(v).t() @ v) / (v.t() @ v)
    return alpha.item()


def compute_dynamic_graph_alpha(data_list: Union[DynamicData, List[Adj]], ignore_disconnected: bool = True):
    """
    Geometric mean of the spectral norms for a dynamic graph

    :param data_list: A dynamic graph or a list of adjacency matrices
    :param ignore_disconnected: Whether to ignore timesteps without edges (default true)
    :return: Geometric mean of alphas
    """
    if isinstance(data_list, DynamicData):
        data_list = [data_list[t].edge_index for t in range(data_list.num_timesteps)]
    alphas = [graph_spectral_norm(edge_index) for edge_index in data_list if
              not ignore_disconnected or edge_index.shape[1] > 0]
    return statistics.geometric_mean(alphas)


def compute_dynamic_weighted_graph_alpha(data: DynamicData, ignore_disconnected: bool = True):
    """
    Geometric mean of the spectral norms for a dynamic weighted graph

    :param data: A dynamic graph or a list of adjacency matrices
    :param ignore_disconnected: Whether to ignore timesteps without edges (default true)
    :return: Geometric mean of alphas
    """
    alphas = [graph_spectral_norm(data[t].edge_index, data[t].edge_weight) for t in range(data.num_timesteps) if
              not ignore_disconnected or data[t].edge_index.shape[1] > 0]
    return statistics.geometric_mean(alphas)


def distance_to_proximity(edge_weight: Union[Tensor, List[Tensor]]):
    """
    Convert edge weights from distances to proximities via max-min normalization

    :param edge_weight: Edge weights (single tensor or temporal list of tensors)
    :return: Proximity edge weights
    """
    if type(edge_weight) is list:
        return [distance_to_proximity(edge_weight_t) for edge_weight_t in edge_weight]
    else:
        max_weight, min_weight = edge_weight.max(), edge_weight.min()
        return 1 - (edge_weight - min_weight) / (max_weight - min_weight)


def to_sparse_adjacency(edge_index: Adj, edge_weight: OptTensor = None,
                        num_nodes: Optional[int] = None) -> SparseTensor:
    """
    Convert edge index and weights to sparse adjacency tensor

    :param edge_index: Edge index
    :param edge_weight: Edges weight (optional)
    :param num_nodes: Number of nodes (optional, but *strongly* advised)
    :return: Sparse adjacency tensor
    """
    return SparseTensor.from_edge_index(edge_index, edge_weight,
                                        (num_nodes, num_nodes) if num_nodes is not None else None)



from typing import Optional, Tuple

import torch
from torch import Size, Tensor

__all__ = ['uniform', 'normal', 'ring', 'orthogonal', 'symmetric', 'antisymmetric', 'diagonal', 'binary', 'ones',
           'zeros', 'rescale_']


def uniform(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
            scale: Optional[float] = None, range: Tuple[float, float] = (-1, 1)) -> Tensor:
    """
    Uniform random tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :param range: Range of uniform distribution (default [-1, +1])
    :return: A random tensor
    """
    W = torch.empty(size).uniform_(*range)
    rescale_(W, rho, sigma, scale)
    return W.data


def normal(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
           scale: Optional[float] = None) -> Tensor:
    """
    Normal random tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.empty(size).normal_(mean=0, std=1)
    rescale_(W, rho, sigma, scale)
    return W.data


def ring(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
         scale: Optional[float] = None) -> Tensor:
    """
    Ring matrix

    See:
    C. Gallicchio & A. Micheli (2020). Ring Reservoir Neural Networks for Graphs.
    In 2020 International Joint Conference on Neural Networks (IJCNN), IEEE.
    https://doi.org/10.1109/IJCNN48605.2020.9206723

    :param size: Size of tensor (must be square)
    :param rho: Spectral radius (equivalent to others)
    :param sigma: Spectral norm (equivalent to others)
    :param scale: Simple rescaling of the matrix (equivalent to others)
    :return: A re-scaled ring matrix
    """
    assert (len(size) == 2) and (size[0] == size[1])
    assert any(arg is not None for arg in [rho, sigma, scale])
    if scale is None:
        scale = rho if sigma is None else sigma
    W = torch.eye(size[0]).roll(1, 0) * scale
    return W.data


def orthogonal(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
               scale: Optional[float] = None) -> Tensor:
    """
    Orthogonal matrix

    See:
    F. Mezzadri (2007). How to Generate Random Matrices from the Classical Compact Groups.
    Notices of the American Mathematical Society, 54(5), pp. 592-604.
    https://www.ams.org/notices/200705/fea-mezzadri-web.pdf

    :param size: Size of tensor (if not square, generates a semi-orthogonal matrix)
    :param rho: Spectral radius (equivalent to others)
    :param sigma: Spectral norm (equivalent to others)
    :param scale: Simple rescaling of the matrix (equivalent to others)
    :return: A re-scaled orthogonal matrix
    """
    assert any(arg is not None for arg in [rho, sigma, scale])
    if scale is None:
        scale = rho if sigma is None else sigma
    W = torch.empty(size)
    torch.nn.init.orthogonal_(W, scale)
    return W.data


def symmetric(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
              scale: Optional[float] = None) -> Tensor:
    """
    Symmetric random tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.empty(size).uniform_(-1, 1)
    W = (W + W.t()) / 2
    rescale_(W, rho, sigma, scale)
    return W.data


def antisymmetric(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
                  scale: Optional[float] = None) -> Tensor:
    """
    Antisymmetric random tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.empty(size).uniform_(-1, 1)
    W = (W - W.t()) / 2
    rescale_(W, rho, sigma, scale)
    return W.data


def diagonal(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
             scale: Optional[float] = None, range: Tuple[float, float] = (-1, 1)) -> Tensor:
    """
    Diagonal uniform random tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :param range: Range of uniform distribution (default [-1, +1])
    :return: A random tensor
    """
    assert size[0] == size[1]
    W = torch.empty(size[0]).uniform_(*range).diag()
    rescale_(W, rho, sigma, scale)
    return W.data


def binary(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
           scale: Optional[float] = None) -> Tensor:
    """
    Binary random tensor (±1)

    Used as input matrix for Ring Reservoirs. See:
    C. Gallicchio & A. Micheli (2020). Ring Reservoir Neural Networks for Graphs.
    In 2020 International Joint Conference on Neural Networks (IJCNN), IEEE.
    https://doi.org/10.1109/IJCNN48605.2020.9206723

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.empty(size).uniform_(-1, 1).sign()
    rescale_(W, rho, sigma, scale)
    return W.data


def ones(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
         scale: Optional[float] = None) -> Tensor:
    """
    Ones tensor

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.ones(size)
    rescale_(W, rho, sigma, scale)
    return W.data


def zeros(size: Size, rho: Optional[float] = None, sigma: Optional[float] = None,
          scale: Optional[float] = None) -> Tensor:
    """
    Zeros tensor

    Rescaling is meaningless in this case.

    :param size: Size of tensor
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: A random tensor
    """
    W = torch.zeros(size)
    return W.data


def rescale_(W: Tensor, rho: Optional[float] = None, sigma: Optional[float] = None,
             scale: Optional[float] = None) -> Tensor:
    """
    Rescale a matrix in-place

    Can either be rescaled according to spectral radius `rho`, spectral norm `sigma`, or `scale`.

    :param W: Matrix to rescale
    :param rho: Spectral radius
    :param sigma: Spectral norm
    :param scale: Simple rescaling of the standard random matrix
    :return: Rescaled matrix
    """
    if rho is not None:
        return W.div_(torch.linalg.eigvals(W).abs().max()).mul_(rho)
    elif sigma is not None:
        return W.div_(torch.linalg.matrix_norm(W, ord=2)).mul_(sigma)
    elif scale is not None:
        return W.mul_(scale)
